"""
Module: schema.py

This module contains Python code for defining and creating a database schema using SQLAlchemy. 
It defines four tables: 'DimCustomer', 'FactPredictions' 'FactPushNotification' and 'FactEmail'.

It also configures a custom logger for informational messages regarding the schema creation.

"""

import logging
import os
from ..logger import CustomFormatter

# Initialize and configure the logger
logger = logging.getLogger(os.path.basename(__file__))
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)

from sqlalchemy import create_engine, Column, Integer, String, Float, DATE, DateTime, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from datetime import datetime

# Define and configure the database engine (Change the connection URL as needed)
engine = create_engine('sqlite:///sa_db.db')

# Create a base class for declarative class definitions
Base = declarative_base()

class Customer(Base):
    """
    Class: Customer

    This class defines the 'DimCustomer' table, which represents customer information.

    Attributes:
    - Customer_ID (int): Primary key for the customer.
    - Age (int): Customer's age.
    - Tenure (int): Customer's tenure with the company.
    - Gender (str): Customer's gender.
    - Income (int): Customer's income.
    - Marital_Status (str): Customer's marital status.
    - Address_ID (int): Customer's address identifier.
    - Education (str): Customer's education level.
    - Retirement (str): Customer's retirement status.
    - Churn (str): Customer's churn status.
    - Region (str): Customer's region.
    - Service_Category (str): Service category.
    - Voice_Included (str): Voice service inclusion status.
    - Internet_Included (str): Internet service inclusion status.
    - Forward_Included (str): Forwarding service inclusion status.
    """
    __tablename__ = "DimCustomer"

    Customer_ID = Column(Integer, primary_key=True)
    Age = Column(Integer)
    Tenure = Column(Integer)
    Gender = Column(String(10))
    Income = Column(Integer)
    Marital_Status = Column(String(15))
    Address_ID = Column(Integer)
    Education = Column(String(45))
    Retirement = Column(String(3))
    Churn = Column(String(3))
    Region = Column(String(15))
    Service_Category = Column(String(15))
    Voice_Included = Column(String(3))
    Internet_Included = Column(String(3))
    Forward_Included = Column(String(3))

class FactPredictions(Base):
    """
    Class: FactPredictions

    This class defines the 'FactPredictions' table, which stores predictive information related to customers.

    Attributes:
    - pred_period (int): Primary key representing the prediction period.
    - customer_ID (int): Foreign key referencing the 'DimCustomer' table.
    - CLV (float): Customer Lifetime Value.
    - Churn_Rate (float): Churn rate.
    - customer (relationship): Establishes a relationship with the 'DimCustomer' table.

    """
    __tablename__ = "FactPredictions"

    pred_period = Column(Integer, primary_key=True)
    customer_ID = Column(Integer, ForeignKey('DimCustomer.Customer_ID'), primary_key=True)
    CLV = Column(Float)
    Churn_Rate = Column(Float)
    customer = relationship("DimCustomer")

class FactPushNotification(Base):
    """
    Class: FactPushNotification

    This class defines the 'FactPushNotification' table, which stores information about the push notifications sent to customers.

    Attributes:
    - sent_date (DateTime): Primary key representing the date the push notification was sent.
    - customer_ID (int): Foreign key referencing the 'DimCustomer' table.
    - success (int): A binary column representing wether the push notification was sent successfully or not.
    - customer (relationship): Establishes a relationship with the 'DimCustomer' table.

    """
    __tablename__ = "FactPushNotification"

    sent_date = Column(DateTime, primary_key=True)
    customer_ID = Column(Integer, ForeignKey('DimCustomer.Customer_ID'), primary_key=True)
    success = Column(Integer)
    customer = relationship("DimCustomer")

class FactEmail(Base):
    """
    Class: FactEmail

    This class defines the 'FactEmail' table, which stores information about the emails sent to customers.

    Attributes:
    - sent_date (DateTime): Primary key representing the date the push notification was sent.
    - customer_ID (int): Foreign key referencing the 'DimCustomer' table.
    - success (int): A binary column representing wether the push notification was sent successfully or not.
    - customer (relationship): Establishes a relationship with the 'DimCustomer' table.

    """
    __tablename__ = "FactEmail"

    sent_date = Column(DateTime, primary_key=True)
    customer_ID = Column(Integer, ForeignKey('DimCustomer.Customer_ID'), primary_key=True)
    success = Column(Integer)
    customer = relationship("DimCustomer")


# Create the tables defined in the schema
Base.metadata.create_all(engine)

# Log a message indicating that the schema has been created
logger.info("Schema Has Been Created")



from fastapi import FastAPI, HTTPException, Query, File, UploadFile, Path
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from datetime import datetime
import sqlite3
import logging
from ..logger import CustomFormatter
from ..database_preparation import SqlHandler
import os
import pandas as pd
from typing import Any, List, Union, Optional
import traceback

app = FastAPI()
    
logger = logging.getLogger(os.path.basename(__file__))
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)

# Creating instances of SqlHandler for each table
dim_customer_handler = SqlHandler(dbname='sa_db', table_name='DimCustomer')
fact_predictions_handler = SqlHandler(dbname='sa_db', table_name='FactPredictions')
fact_push_notification_handler = SqlHandler(dbname='sa_db', table_name='FactPushNotification')
fact_email_handler = SqlHandler(dbname='sa_db', table_name='FactEmail')

# Defining functions to open connections to the databases
def get_dim_customer_db():
    return dim_customer_handler.cnxn

def get_fact_predictions_db():
    return fact_predictions_handler.cnxn

def get_fact_push_notification_db():
    return fact_push_notification_handler.cnxn

def get_fact_email_db():
    return fact_email_handler.cnxn

@app.on_event("shutdown")
async def shutdown_event():
    # Closing the database connections on shutdown
    dim_customer_handler.close_cnxn()
    fact_predictions_handler.close_cnxn()
    fact_push_notification_handler.close_cnxn()
    fact_email_handler.close_cnxn()

@app.get("/")
async def root():
    return {
        """
            Initializing API for Survival Analysis project with selecting data from database, inserting data, and updating data. It now contains also several endpoints that help to identify top customers with the highest churn rate, customers with the highest/lowest CLV. Add /docs to see available endpoints.
        """
    }

def read_csv(file: UploadFile = File(...)):
    content = pd.read_csv(file.file)
    return content

def validate_sent_date(sent_date):
    try:
        # Trying to parse the send_date as a datetime object
        datetime.strptime(sent_date, '%d/%m/%Y')
        return True
    except ValueError:
        return False

@app.put("/populate_fact_push_notification")
async def populate_fact_push_notification(file: UploadFile = File(...)):
    try:
        # Reading the CSV file
        data = read_csv(file)
        
        # Validating sent_date column
        if 'sent_date' in data.columns and not all(data['sent_date'].apply(validate_sent_date)):
            raise HTTPException(status_code=400, detail="Invalid or missing values in the sent_date column.")
        
        # Inserting data into FactPushNotification table
        fact_push_notification_handler.insert_many(data)

        return JSONResponse(content=jsonable_encoder({"message": "Data loaded successfully"}), status_code=200)

    except Exception as e:
        logger.error(f"Failed to populate FactPushNotification: {str(e)}")
        raise HTTPException(status_code=500, detail=f"Failed to populate FactPushNotification: {str(e)}")
    
@app.put("/populate_fact_email")
async def populate_fact_email(file: UploadFile = File(...)):
    try:
        # Reading the CSV file
        data = read_csv(file)
        
        # Validating sent_date column
        if 'sent_date' in data.columns and not all(data['sent_date'].apply(validate_sent_date)):
            raise HTTPException(status_code=400, detail="Invalid or missing values in the sent_date column.")
        
        # Inserting data into FactEmail table
        fact_email_handler.insert_many(data)

        return JSONResponse(content=jsonable_encoder({"message": "Data loaded successfully"}), status_code=200)

    except Exception as e:
        logger.error(f"Failed to populate FactEmail: {str(e)}")
        logger.error(traceback.format_exc())
        raise HTTPException(status_code=500, detail=f"Failed to populate FactEmail: {str(e)}")

    
## Adding endpoints for different scenarios

@app.get("/get_top_churn_clv_customers")
async def get_top_churn_clv_customers(
    pred_period: int = Query(..., description="Prediction period (1-12)"),
    top_percentage: int = Query(10, description="Top percentage of customers to select (e.g., 10 for top 10%)", ge=1, le=100)
):
    try:
        handler = fact_predictions_handler

        # Calling the from_sql_to_pandas function to get the data for FactPredictions
        fact_data = handler.from_sql_to_pandas(chunksize=1000, id_value='customer_ID')

        # Filtering the data for the specified prediction period
        selected_data = fact_data[fact_data['pred_period'] == pred_period]

        # Sorting the data first by Churn_Rate and then by CLV in descending order
        sorted_data = selected_data.sort_values(by=['Churn_Rate', 'CLV'], ascending=[False, False])

        # Calculating the number of customers to select based on the specified top percentage
        total_customers = len(sorted_data)
        top_count = int(top_percentage / 100 * total_customers)

        # Selecting the top percentage of customers with the highest churn rate and then CLV
        top_churn_customers = sorted_data.head(top_count)

        if top_churn_customers.empty:
            return {"message": "No customers found"}

        # Returning pred_period, customer_id, churn_rate, and clv without merging
        result_data = top_churn_customers[['pred_period', 'customer_ID', 'Churn_Rate', 'CLV']].to_dict(orient='records')

        return result_data
    except Exception as e:
        logger.error(f"Failed to get top churn customers: {str(e)}")
        raise HTTPException(status_code=500, detail=f"Failed to get top churn customers: {str(e)}")


@app.get("/get_top_clv_customers")
async def get_top_clv_customers(
    top_percentage: int = Query(..., description="Percentage of customers to select (e.g., 20 for top 20%)", ge=1, le=100),
    pred_period: int = Query(12, description="Prediction period (1-12)")
):
    try:
        handler = fact_predictions_handler

        # Calling the from_sql_to_pandas function to get the data for FactPredictions
        fact_data = handler.from_sql_to_pandas(chunksize=1000, id_value='customer_ID')

        # Filtering the data for the specified prediction period
        selected_data = fact_data[fact_data['pred_period'] == pred_period]

        # Sorting the data by CLV in descending order
        sorted_data = selected_data.sort_values(by='CLV', ascending=False)

        # Calculating the number of customers to select based on the specified top percentage
        total_customers = len(sorted_data)
        selected_count = int(top_percentage / 100 * total_customers)

        # Selecting the top percentage of customers
        selected_customers = sorted_data.head(selected_count)

        if selected_customers.empty:
            return {"message": "No customers found"}

        # Returning only customer_id and pred_period without merging
        result_data = selected_customers[['customer_ID', 'pred_period']].to_dict(orient='records')

        return {"selected_customers": result_data}
    except Exception as e:
        logger.error(f"Failed to get top CLV customers: {str(e)}")
        raise HTTPException(status_code=500, detail=f"Failed to get top CLV customers: {str(e)}")






import pandas as pd
import os
import logging
from ..logger import CustomFormatter
from lifelines import WeibullAFTFitter, LogNormalAFTFitter, LogLogisticAFTFitter
from lifelines.fitters import ParametricRegressionFitter
from autograd import numpy as np

# Initialize and configure the logger
logger = logging.getLogger(os.path.basename(__file__))
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)


class ExponentialAFTFitter(ParametricRegressionFitter):
    '''
    This is a class for implementing an Exponential AFT Fitter Model.
    '''
    # this class property is necessary, and should always be a non-empty list of strings.
    _fitted_parameter_names = ['lambda_']

    def _cumulative_hazard(self, params, t, Xs):
        # params is a dictionary that maps unknown parameters to a numpy vector.
        # Xs is a dictionary that maps unknown parameters to a numpy 2d array
        beta = params['lambda_']
        X = Xs['lambda_']
        lambda_ = np.exp(np.dot(X, beta))
        return t / lambda_

class AFTModelSelector:
    """
    A class for selecting the best AFT (Accelerated Failure Time) model among Weibull, Exponential,
    Log-Normal, and Log-Logistic models based on AIC, and generating churn rate and customer lifetime value (CLV) 
    predictions for a specified number of time periods.

    Parameters:
    - data (pd.DataFrame): The input DataFrame containing survival data.
    - primary_col(str): The column name in the DataFrame representing the primary key.
    - duration_col (str): The column name in the DataFrame representing the duration or time-to-event.
    - event_col (str): The column name in the DataFrame representing the event indicator.

    Attributes:
    - data (pd.DataFrame): The input DataFrame containing survival data.
    - primary(str): The column name in the DataFrame representing the primary key.
    - duration_col (str): The column name in the DataFrame representing the duration or time-to-event.
    - event_col (str): The column name in the DataFrame representing the event indicator.
    - aft_model (lifelines.Fitter): The selected AFT model based on AIC.
    - predictions_df (pd.DataFrame): DataFrame containing churn and CLV predictions for a specified number of time periods.
    """
    
    def __init__(self, data: pd.DataFrame , primary_col:str,  duration_col : str, event_col: str):
        self.data = data
        self.primary = primary_col
        self.duration_col = duration_col
        self.event_col = event_col
        self.aft_model = None
        self.predictions_df = None

            
            
    def select_best_model(self):
        """
        Selects the best AFT model among Weibull, Exponential, Log-Normal, and Log-Logistic models based on AIC.
        Stores the selected model in the 'aft_model' attribute.
        """
        models = {
            'Weibull': WeibullAFTFitter(),
            'Exponential': ExponentialAFTFitter(),
            'LogNormal': LogNormalAFTFitter(),
            'LogLogistic': LogLogisticAFTFitter(),
        }

        best_aic = float('inf')
        best_model = None

        # Handle zero values in the duration column
        self.data[self.duration_col] = self.data[self.duration_col].replace(0, 0.0001)

        for model_name, model in models.items():
            model.fit(self.data, duration_col= self.duration_col, event_col= self.event_col)

            aic = model.AIC_
            logger.info(f"{model_name} AIC: {aic}")
    
            if aic < best_aic:
                best_aic = aic
                best_model = model_name

        logger.warning(f"\nBest Model: {best_model} with AIC: {best_aic}")
        self.aft_model = models[best_model]


    def fit_and_predict(self, n_time_periods: int):
        """
        Fits the selected AFT model and generates churn predictions for a specified number of time periods.
        Stores the predictions in the 'predictions_df' attribute.

        Parameters:
        - n_time_periods (int): The number of time periods for which predictions should be generated.
            
        Returns:
        - str: A message indicating the model ran successfully. 
        """
        if self.aft_model is None:
            logger.warning("Please run select_best_model() first.")
            return

        # Handle zero values in the duration column
        self.data[self.duration_col] = self.data[self.duration_col].replace(0, 0.0001)

        predictions_df_list = []

        for time_period in range(1, n_time_periods + 1):
            customer_data = pd.DataFrame({
                'customer_id': self.data[self.primary],
                'pred_period': time_period
            })

            # Generate survival predictions 
            predictions = self.aft_model.predict_survival_function(self.data, times=[time_period])

            #obtaining churn predictions
            churn = round(1 - predictions, 5)
            # Convert predictions to a DataFrame
            predictions_df = pd.DataFrame(churn.T.values, columns=['churn_rate'])

            # Concatenate customer_id and time_period with predictions DataFrame
            result_df = pd.concat([customer_data, predictions_df], axis=1)

            # Append to the list
            predictions_df_list.append(result_df)

        # Concatenate all predictions into a single DataFrame
        self.predictions_df = pd.concat(predictions_df_list, ignore_index=True)
        logger.info("The AFT model was run successfully.")

    def calculate_clv(self, MM=1300, r=0.1):
        """
        Calculates Customer Lifetime Value (CLV) for each customer in 'predictions_df' attribute 
        and updates the dataframe to include CLV predictions.

        Parameters:
        - MM (float): A constant representing the monetary value.
        - r (float): The periodic interest rate for discounting.
                
        Returns:
        - str: A message indicating the 'predictions_df' attribute was updated successfully. 
        """
        
        predictions_dfs = []  # List to store individual CLV prediction DataFrames
        if self.predictions_df is None:
            logger.warning("Please run fit_and_predict() first.")
            return
        
        #Making the data long format
        data_clv = self.predictions_df.pivot(index='customer_id', columns='pred_period', values='churn_rate')
        #Calculating the Survival Rates from Churn Rates 
        data_clv = 1 - data_clv

        # Iterating over an increasing number of columns
        for i in range(1, len(data_clv.columns) + 1):
            # Selecting the first i columns
            subset_data = data_clv.iloc[:, :i]
    
            # Calculating CLV 
            data_clv1 = subset_data.copy()
            sequence = list(range(1, len(data_clv1.columns) + 1))

            for num in sequence:
                data_clv1.iloc[:, num - 1] /= (1 + r/12) ** (num - 1)

            data_clv1['CLV'] = MM * data_clv1.sum(axis=1)
            
            predictions_dfs.append(data_clv1['CLV'])

        clv_prediction = pd.concat(predictions_dfs, axis=1)
        clv_prediction.columns = range(1, len(predictions_dfs) + 1)

        #returning to original data format and saving the predictions
        clv_prediction = clv_prediction.reset_index()
        clv_prediction = pd.melt(clv_prediction, id_vars=['customer_id'], var_name='pred_period', value_name='CLV')

        #Combining the results and updating the predictions dataframe
        self.predictions_df = pd.merge(self.predictions_df, clv_prediction, on=['customer_id','pred_period'], how='left')
        logger.info("The CLV predictions were added successfully.")
        




import logging

class CustomFormatter(logging.Formatter):
    
    """ 
        Custom formatter for Informative Logging
    """
    grey = "\x1b[38;20m"
    violet="\x1b[38;5;183m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(name)s - %(funcName)s - %(levelname)s - (%(message)s) - line: %(lineno)d"
    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: violet + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }
    def format(self, record):
        """

        Parameters
        ----------
        record : takes the record
            

        Returns
        -------
        returns formated(colored) output
        """
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


if __name__=='__main__':
    # Just en example
    import os 
    
    logger = logging.getLogger(os.path.basename(__file__))
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)

    logger.debug("debug message")
    logger.info("Warning: Email has not been sent......")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")


from setuptools import setup, find_packages

setup(
    author='Anna Manasyan, Anna Shaljyan, Ela Khachatryan, Sergey Tovmasyan',
    description='Survival Analysis: Customer Churn and CLV Prediction ',
    name='survival_analysis',
    version='0.0.1',
    packages=find_packages(include=['survival_analysis','survival_analysis.*']),
    include_package_data= True
)


"""
Module: sql_interactions.py

This module defines a Python class called 'SqlHandler' for interacting with SQLite databases. The class allows for various operations on SQLite databases, such as connecting, inserting data, retrieving data, and more.

Attributes:
- dbname (str): The name of the SQLite database.
- table_name (str): The name of the table within the database.

Methods:

- __init__(self, dbname: str, table_name: str) -> None:
    Constructor for the 'SqlHandler' class. Initializes a connection to an SQLite database and specifies the target table.

- close_cnxn(self) -> None:
    Closes the SQLite database connection.

- insert_one(self, data) -> str:
    Inserts a single row to the database. 

- get_table_columns(self) -> list:
    Retrieves a list of column names for the specified table.

- truncate_table(self) -> None:
    Truncates the specified table, removing all its data.

- drop_table(self) -> None:
    Deletes the specified table from the database.

- insert_many(self, df: pd.DataFrame) -> str:
    Inserts data from a pandas DataFrame into the specified table, mapping columns to their respective database columns.

- from_sql_to_pandas(self, chunksize: int, id_value: str) -> pd.DataFrame:
    Retrieves data from the database and loads it into a pandas DataFrame in chunks of a specified size.

- update_table(self, set_dict, cond_dict) -> str:
    Update rows in a database table based on the set and where conditions provided in dictionaries.


"""

import sqlite3
import logging
import pandas as pd
import numpy as np
import os
from ..logger import CustomFormatter

# Initialize and configure the logger
logger = logging.getLogger(os.path.basename(__file__))
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)

class SqlHandler:

    def __init__(self, dbname: str, table_name: str) -> None:
        """
        Constructor for the SqlHandler class.

        Parameters:
        - dbname (str): The name of the SQLite database.
        - table_name (str): The name of the table within the database.
        """
        self.cnxn = sqlite3.connect(f'{dbname}.db')
        self.cursor = self.cnxn.cursor()
        self.dbname = dbname
        self.table_name = table_name

    def close_cnxn(self) -> None:
        """
        Closes the SQLite database connection.
        
        """
        logger.info('Committing the changes')
        self.cnxn.close()
        logger.info('The connection has been closed')

    def insert_one(self, data: dict) -> str:
        """
        Inserts a single record from a dictionary into the specified table, mapping keys to their respective database columns.

        Parameters:
        - data (dict): A dictionary containing data to be inserted.

        Returns:
        - str: A message indicating that the data has been loaded.
        """

        #getting the DB columns

        columns = self.get_table_columns()
        sql_column_names = [col.lower() for col in columns]

        # Convert keys to lowercase
        data = {key.lower(): value for key, value in data.items()}

        # Filter the dictionary to keep only keys that match the database column names
        filtered_data = {key: value for key, value in data.items() if key in sql_column_names}
        
        # Prepare the values to be inserted
        row_values = list(filtered_data.values())
        columns = list(filtered_data.keys())
        
        ncolumns = ['?' for _ in columns]
        cols = ', '.join(columns)
        params = ', '.join(ncolumns)

        logger.info(f'Insert structure: colnames: {cols} params: {params}')

        query = f"""INSERT INTO {self.table_name} ({cols}) VALUES ({params});"""

        self.cursor.execute(query, row_values)
    
        try:
            logger.info(self.cursor.messages)
        except:
            pass

        self.cnxn.commit()
        logger.warning('The data is loaded')


    def get_table_columns(self) -> list:
        """
        Retrieves a list of column names for the specified table.

        Returns:
        - column_names (list): List of column names in the table.
        """
        self.cursor.execute(f"PRAGMA table_info({self.table_name});")
        columns = self.cursor.fetchall()
        column_names = [col[1] for col in columns]
        logger.info(f'The list of columns: {column_names}')
        return column_names
    
    def truncate_table(self) -> None:
        """
        Truncates the specified table, removing all its data.
        """
        query = f"DELETE FROM {self.table_name};"
        self.cursor.execute(query)
        self.cnxn.commit()
        logging.info(f'The {self.table_name} is truncated')
        self.cursor.close()

    def drop_table(self) -> None:
        """
        Deletes the specified table from the database.
        """
        query = f"DROP TABLE IF EXISTS {self.table_name};"
        logging.info(query)
        self.cursor.execute(query)
        self.cnxn.commit()
        logging.info(f"Table '{self.table_name}' deleted.")
        logger.debug('Using drop table function')

    def insert_many(self, df: pd.DataFrame) -> str:
        """
        Inserts data from a pandas DataFrame into the specified table, mapping columns to their respective database columns.

        Parameters:
        - df (pd.DataFrame): The DataFrame containing data to be inserted.

        Returns:
        - str: A message indicating that the data has been loaded.
        """
        df = df.replace(np.nan, None)  # For handling NULLS
        df.rename(columns=lambda x: x.lower(), inplace=True)
        columns = list(df.columns)
        logger.info(f'BEFORE the column intersection: {columns}')
        sql_column_names = [i.lower() for i in self.get_table_columns()]
        columns = list(set(columns) & set(sql_column_names))
        logger.info(f'AFTER the column intersection: {columns}')
        ncolumns = list(len(columns) * '?')
        data_to_insert = df.loc[:, columns]
        values = [tuple(i) for i in data_to_insert.values]
        logger.info(f'The shape of the table which is going to be imported {data_to_insert.shape}')
        if len(columns) > 1:
            cols, params = ', '.join(columns), ', '.join(ncolumns)
        else:
            cols, params = columns[0], ncolumns[0]
        logger.info(f'Insert structure: colnames: {cols} params: {params}')
        logger.info(values[0])
        query = f"""INSERT INTO {self.table_name} ({cols}) VALUES ({params});"""
        logger.info(f'QUERY: {query}')
        self.cursor.executemany(query, values)
        try:
            for i in self.cursor.messages:
                logger.info(i)
        except:
            pass
        self.cnxn.commit()
        logger.warning('The data is loaded')

    def from_sql_to_pandas(self, chunksize: int, id_value: str) -> pd.DataFrame:
        """
        Retrieves data from the database and loads it into a pandas DataFrame in chunks of a specified size.

        Parameters:
        - chunksize (int): The chunksize for data extraction.
        - id_value (str): The values by which the data should be sorted.

        Returns:
        - pd.DataFrame: A DataFrame containing the retrieved data.
        """
        offset = 0
        dfs = []
        while True:
            query = f"""
            SELECT * FROM {self.table_name}
                ORDER BY {id_value} 
                LIMIT {chunksize}
                OFFSET {offset};
            """
            data = pd.read_sql_query(query, self.cnxn)
            logger.info(f'The shape of the chunk: {data.shape}')
            dfs.append(data)
            offset += chunksize
            if len(dfs[-1]) < chunksize:
                logger.warning('Loading the data from SQL is finished')
                logger.debug('Connection is closed')
                break
        df = pd.concat(dfs)
        return df

   
    def update_table(self, set_dict: dict, cond_dict: dict) -> str:
        """
        Update rows in a database table based on the set and where conditions provided in dictionaries.

        Parameters:
        - set_dict (dict): A dictionary with column names as keys and new values as values for the SET clause.
        - cond_dict (dict): A dictionary with column names as keys and conditions as values for the WHERE clause.

        Returns:
        - str: A message indicating that the data has been updated.
        """
        
        #getting the db column names
        columns = self.get_table_columns()
        sql_column_names = [col.lower() for col in columns]

        # Convert keys to lowercase
        set_dict = {key.lower(): value for key, value in set_dict.items()}
        cond_dict = {key.lower(): value for key, value in cond_dict.items()}

        # Filter the dictionary to keep only keys that match the database column names
        set_dict = {key: value for key, value in set_dict.items() if key in sql_column_names}
        cond_dict = {key: value for key, value in cond_dict.items() if key in sql_column_names}    
        
        try:
            # Generate the SET clause
            set_clause = ', '.join([f"{col} = ?" for col in set_dict.keys()])
            set_values = list(set_dict.values())
            logger.info(f'Set Clause Structure: {set_clause}')

            # Generate the WHERE clause
            where_clause = ' AND '.join([f"{col} = ?" for col in cond_dict.keys()])
            where_values = list(cond_dict.values())  # Add values for the WHERE clause
            logger.info(f'Where Clause Structure: {where_clause}')

            # Build the SQL query
            query = f"""
                UPDATE {self.table_name}
                SET {set_clause}
                WHERE {where_clause};
                    """
            logger.info(f'Generated SQL query: {query}')  # Log the generated query
            self.cursor.execute(query, set_values + where_values)  # Combine SET and WHERE values

            self.cnxn.commit()
            logger.warning(f'The table {self.table_name} is updated.')
        except Exception as e:
            logger.warning(f"Error updating rows: {e}")

            





import pandas as pd

""""
THIS IS A MODULE FOR UTILITY FUNCTIONS
"""

def format_dataframe(df):
    """
    Converts categorical variables in a DataFrame to binary columns using one-hot encoding and makes sure that numeric variables are of correct type.

    Parameters:
        df (pd.DataFrame): The input DataFrame.

    Returns:
        pd.DataFrame: The DataFrame with categorical variables converted to binary columns.
    """
    # Identify string variables
    string_variables = df.select_dtypes(include=['object']).columns

    #Exclude columns where all values can be converted to numeric
    numeric_variables = df.apply(pd.to_numeric, errors='coerce').notna().all()
    string_variables = string_variables.difference(numeric_variables[numeric_variables].index)
    df.reset_index(drop=True, inplace=True)
    
    # Filter DataFrame to keep only string variables
    string_df = df[string_variables]

    # Dummify the string variables
    dummies_df = pd.get_dummies(string_df, columns=string_variables, prefix=string_variables,  drop_first=True)
    dummies_df = dummies_df.astype(int)

    # Concatenate the dummified variables with the original DataFrame
    output_df = pd.concat([df, dummies_df], axis=1)

    # Convert true numeric values to appropriate numeric type
    for column in numeric_variables[numeric_variables].index:
        output_df[column] = pd.to_numeric(df[column], errors='coerce')

    # Drop the original string variables
    output_df = output_df.drop(columns=string_variables)

    return output_df

